from airflow.models.operator import Operator
from airflow.providers.ssh.hooks.ssh import SSHHook
from airflow.providers.ssh.operators.ssh import SSHOperator
from shlex import quote
from typing import Dict, List, Optional, Union

from airflow_supervisor.config import SupervisorAirflowConfiguration

from .common import SupervisorTaskStep
from .local import Supervisor

__all__ = ("SupervisorSSH",)


class SupervisorSSH(Supervisor):
    # Mimic SSH Operator: https://airflow.apache.org/docs/apache-airflow-providers-ssh/stable/_api/airflow/providers/ssh/operators/ssh/index.html
    def __init__(
        self,
        supervisor_cfg: SupervisorAirflowConfiguration,
        command_prefix: str = "",
        command_noescape: str = "",
        ssh_hook: Optional[SSHHook] = None,
        ssh_conn_id: Optional[str] = None,
        remote_host: Optional[str] = None,
        conn_timeout: Optional[int] = None,
        cmd_timeout: Optional[int] = None,
        environment: Optional[dict] = None,
        get_pty: Optional[bool] = None,
        banner_timeout: Optional[float] = None,
        skip_on_exit_code: Optional[Union[int, List[int]]] = None,
        **kwargs,
    ):
        self._command_prefix = command_prefix
        self._command_noescape = command_noescape
        self._ssh_operator_kwargs = {}
        if ssh_hook:
            self._ssh_operator_kwargs["ssh_hook"] = ssh_hook
        if ssh_conn_id:
            self._ssh_operator_kwargs["ssh_conn_id"] = ssh_conn_id
        if remote_host:
            self._ssh_operator_kwargs["remote_host"] = remote_host
        if conn_timeout:
            self._ssh_operator_kwargs["conn_timeout"] = conn_timeout
        if cmd_timeout:
            self._ssh_operator_kwargs["cmd_timeout"] = cmd_timeout
        if environment:
            self._ssh_operator_kwargs["environment"] = environment
        if get_pty:
            self._ssh_operator_kwargs["get_pty"] = get_pty
        if banner_timeout:
            self._ssh_operator_kwargs["banner_timeout"] = banner_timeout
        if skip_on_exit_code:
            self._ssh_operator_kwargs["skip_on_exit_code"] = skip_on_exit_code
        super().__init__(supervisor_cfg=supervisor_cfg, **kwargs)

    def get_base_operator_kwargs(self) -> Dict:
        return dict(dag=self, **self._ssh_operator_kwargs)

    def get_step_kwargs(self, step: SupervisorTaskStep) -> Dict:
        if step == "configure-supervisor":
            # TODO
            return dict(
                command=f"""
{self._command_noescape}
{quote(self._command_prefix)}
_airflow_supervisor_command {step} {self._supervisor_cfg.model_dump_json()}
"""
            )
        elif step in ("start-supervisor", "stop-supervisor", "unconfigure-supervisor", "force-kill"):
            # must be done via SSH
            return dict(
                command=f"""
{self._command_noescape}
{quote(self._command_prefix)}
_airflow_supervisor_command {step} --cfg {self._supervisor_cfg._pydantic_path}
"""
            )
        else:
            # can be done via XMLRPC API
            return super().get_step_kwargs(step=step)

    def get_step_operator(self, step: SupervisorTaskStep) -> Operator:
        if step in (
            "configure-supervisor",
            "start-supervisor",
            "stop-supervisor",
            "unconfigure-supervisor",
            "force-kill",
        ):
            return SSHOperator(
                **{"task_id": f"{self.dag_id}-{step}", **self.get_base_operator_kwargs(), **self.get_step_kwargs(step)}
            )
        else:
            # can be done via XMLRPC API
            return super().get_step_operator(step=step)



from airflow.exceptions import AirflowSkipException
from airflow.models.dag import DAG
from airflow.models.operator import Operator
from airflow.operators.python import BranchPythonOperator, PythonOperator
from airflow.operators.trigger_dagrun import TriggerDagRunOperator
from airflow.sensors.python import PythonSensor
from typing import Dict

from airflow_supervisor.client import SupervisorRemoteXMLRPCClient
from airflow_supervisor.config import SupervisorAirflowConfiguration

from .common import SupervisorTaskStep, fail_, pass_, skip_

__all__ = ("Supervisor",)


class Supervisor(DAG):
    _supervisor_cfg: SupervisorAirflowConfiguration
    _supervisor_kill: DAG
    _supervisor_xmlrpc_client: SupervisorRemoteXMLRPCClient

    def __init__(self, supervisor_cfg: SupervisorAirflowConfiguration, **kwargs):
        # store config
        self._supervisor_cfg = supervisor_cfg
        self._supervisor_xmlrpc_client = kwargs.pop(
            "supervisor_xmlrpc_client", SupervisorRemoteXMLRPCClient(self._supervisor_cfg)
        )

        # setup role and tweak dag id
        if "dag_id" not in kwargs:
            kwargs["dag_id"] = list(self._supervisor_cfg.program.keys())[0]

        # override dag kwargs that dont make sense
        kwargs["catchup"] = False
        kwargs["concurrency"] = 1
        kwargs["max_active_tasks"] = 1
        kwargs["max_active_runs"] = 1

        # init with base DAG
        super().__init__(**kwargs)

        # initialize tasks
        self.initialize_tasks()

        # configure graph
        trigger_self_good = TriggerDagRunOperator(
            task_id=f"{self.dag_id}-trigger-loop",
            trigger_dag_id=self.dag_id,
        )
        trigger_self_bad = TriggerDagRunOperator(
            task_id=f"{self.dag_id}-trigger-redo",
            trigger_dag_id=self.dag_id,
        )
        fail_task = PythonOperator(task_id=f"{self.dag_id}-fail", python_callable=fail_)
        skip_task = PythonOperator(task_id=f"{self.dag_id}-skip", python_callable=skip_)
        pass_task = PythonOperator(task_id=f"{self.dag_id}-pass", python_callable=pass_, trigger_rule="one_success")

        # TODO check if we're past the dag's end time
        _branch_choices = {
            # "running": trigger_self_good.task_id,
            # "ok": trigger_self_good.task_id,
            "done": self.stop_programs.task_id,
            "running": trigger_self_good.task_id,
            "ok": trigger_self_good.task_id,
            "fail": self.restart_programs.task_id,
        }

        def _choose_branch(dag_id=self.dag_id, branch_choices=_branch_choices.copy(), **kwargs):
            task_instance = kwargs["task_instance"]
            check_program_result = task_instance.xcom_pull(key="return_value", task_ids=f"{dag_id}-check-programs")
            ret = branch_choices.get(check_program_result, None)
            if ret is None:
                raise AirflowSkipException
            return ret

        check_programs_decide = BranchPythonOperator(
            task_id=f"{self.dag_id}-check-programs-decide",
            python_callable=_choose_branch,
            provide_context=True,
            trigger_rule="all_done",
        )

        self.configure_supervisor >> self.start_supervisor >> self.start_programs >> self.check_programs
        self.check_programs >> check_programs_decide
        # fail, restart
        check_programs_decide >> self.restart_programs >> trigger_self_bad >> fail_task
        # pass, finish
        check_programs_decide >> self.stop_programs >> self.stop_supervisor >> self.unconfigure_supervisor >> pass_task
        # loop
        check_programs_decide >> trigger_self_good >> pass_task

        # TODO make helper dag
        self._force_kill = self.get_step_operator("force-kill")
        # Default non running
        skip_task >> self._force_kill

    def initialize_tasks(self):
        # tasks
        self._configure_supervisor = self.get_step_operator(step="configure-supervisor")
        self._start_supervisor = self.get_step_operator(step="start-supervisor")
        self._start_programs = self.get_step_operator("start-programs")
        self._stop_programs = self.get_step_operator("stop-programs")
        self._restart_programs = self.get_step_operator("restart-programs")
        self._stop_supervisor = self.get_step_operator("stop-supervisor")
        self._unconfigure_supervisor = self.get_step_operator("unconfigure-supervisor")

        # TODO check programs should be sensor
        self._check_programs = self.get_step_operator("check-programs")

    @property
    def configure_supervisor(self) -> Operator:
        return self._configure_supervisor

    @property
    def start_supervisor(self) -> Operator:
        return self._start_supervisor

    @property
    def start_programs(self) -> Operator:
        return self._start_programs

    @property
    def check_programs(self) -> Operator:
        return self._check_programs

    @property
    def stop_programs(self) -> Operator:
        return self._stop_programs

    @property
    def restart_programs(self) -> Operator:
        return self._restart_programs

    @property
    def stop_supervisor(self) -> Operator:
        return self._stop_supervisor

    @property
    def unconfigure_supervisor(self) -> Operator:
        return self._unconfigure_supervisor

    @property
    def supervisor_client(self) -> SupervisorRemoteXMLRPCClient:
        return SupervisorRemoteXMLRPCClient(self._supervisor_cfg)

    def get_base_operator_kwargs(self) -> Dict:
        return dict(dag=self)

    def get_step_kwargs(self, step: SupervisorTaskStep) -> Dict:
        if step == "configure-supervisor":
            from .commands import write_supervisor_config

            return dict(
                python_callable=lambda: write_supervisor_config(self._supervisor_cfg, _exit=False), do_xcom_push=True
            )
        elif step == "start-supervisor":
            from .commands import start_supervisor

            return dict(
                python_callable=lambda: start_supervisor(self._supervisor_cfg._pydantic_path, _exit=False),
                do_xcom_push=True,
            )
        elif step == "start-programs":
            from .commands import start_programs

            return dict(python_callable=lambda: start_programs(self._supervisor_cfg, _exit=False), do_xcom_push=True)
        elif step == "stop-programs":
            from .commands import stop_programs

            return dict(python_callable=lambda: stop_programs(self._supervisor_cfg, _exit=False), do_xcom_push=True)
        elif step == "check-programs":
            from .commands import check_programs

            def _check_programs(cfg=self._supervisor_cfg, **kwargs):
                task_instance = kwargs["task_instance"]
                # TODO formalize
                if check_programs(cfg, check_done=True, _exit=False):
                    task_instance.xcom_push(key="return_value", value="done")

                    # finish
                    return True

                if check_programs(cfg, check_running=True, _exit=False):
                    task_instance.xcom_push(key="return_value", value="running")
                    return False

                if check_programs(cfg, _exit=False):
                    task_instance.xcom_push(key="return_value", value="ok")
                    return False

                task_instance.xcom_push(key="return_value", value="fail")
                return True

            return dict(python_callable=_check_programs, do_xcom_push=True)
        elif step == "restart-programs":
            from .commands import restart_programs

            return dict(python_callable=lambda: restart_programs(self._supervisor_cfg, _exit=False), do_xcom_push=True)
        elif step == "stop-supervisor":
            from .commands import stop_supervisor

            return dict(python_callable=lambda: stop_supervisor(self._supervisor_cfg, _exit=False), do_xcom_push=True)
        elif step == "unconfigure-supervisor":
            from .commands import remove_supervisor_config

            return dict(
                python_callable=lambda: remove_supervisor_config(self._supervisor_cfg, _exit=False), do_xcom_push=True
            )
        elif step == "force-kill":
            from .commands import kill_supervisor

            return dict(python_callable=lambda: kill_supervisor(self._supervisor_cfg, _exit=False), do_xcom_push=True)
        raise NotImplementedError

    def get_step_operator(self, step: SupervisorTaskStep) -> Operator:
        if step == "check-programs":
            return PythonSensor(
                **{
                    "task_id": f"{self.dag_id}-{step}",
                    "poke_interval": self._supervisor_cfg.check_interval.total_seconds(),
                    "timeout": self._supervisor_cfg.check_timeout.total_seconds(),
                    "mode": "reschedule",
                    **self.get_base_operator_kwargs(),
                    **self.get_step_kwargs(step),
                }
            )
        return PythonOperator(
            **{"task_id": f"{self.dag_id}-{step}", **self.get_base_operator_kwargs(), **self.get_step_kwargs(step)}
        )



from pydantic import Field
from typing import Any, Dict, Optional

from .base import _BaseCfgModel

__all__ = ("RpcInterfaceConfiguration",)


class RpcInterfaceConfiguration(_BaseCfgModel):
    def to_cfg(self, key: str) -> str:
        # Overload to require key
        return (
            super()
            .to_cfg(key=key)
            .replace("[rpc_interface", "[rpcinterface")
            .replace("rpcinterface_factory=", "supervisor.rpcinterface_factory=")
        )

    rpcinterface_factory: str = Field(
        default="supervisor.rpcinterface:make_main_rpcinterface",
        description="pkg_resources “entry point” dotted name to your RPC interface’s factory function.",
    )
    kwargs: Optional[Dict[str, Any]] = Field(default=None)  # TODO



from subprocess import check_call

from airflow_supervisor.airflow.commands import start_supervisor, stop_supervisor, write_supervisor_config
from airflow_supervisor.config import SupervisorAirflowConfiguration


def test_command():
    assert check_call(["_airflow_supervisor_command", "--help"]) == 0


def test_write(supervisor_airflow_configuration: SupervisorAirflowConfiguration):
    json = supervisor_airflow_configuration.model_dump_json()
    assert write_supervisor_config(json, _exit=False)
    assert supervisor_airflow_configuration._pydantic_path.read_text().strip() == json
    supervisor_airflow_configuration.rmdir()


def test_start_stop(supervisor_airflow_configuration: SupervisorAirflowConfiguration):
    json = supervisor_airflow_configuration.model_dump_json()
    assert write_supervisor_config(json, _exit=False)
    assert supervisor_airflow_configuration._pydantic_path.read_text().strip() == json
    assert start_supervisor(supervisor_airflow_configuration._pydantic_path, _exit=False)
    assert stop_supervisor(supervisor_airflow_configuration._pydantic_path, _exit=False)
    supervisor_airflow_configuration.rmdir()



import os
from datetime import datetime, timedelta, timezone
from hydra import compose, initialize_config_dir
from hydra.utils import instantiate
from logging import getLogger
from pathlib import Path
from pydantic import BaseModel, Field, PrivateAttr, model_validator
from shutil import rmtree
from signal import SIGKILL, SIGTERM
from subprocess import Popen
from tempfile import gettempdir
from typing import Dict, Optional

from ..exceptions import ConfigNotFoundError
from ..utils import _get_calling_dag
from .airflow import AirflowConfiguration
from .eventlistener import EventListenerConfiguration
from .fcgiprogram import FcgiProgramConfiguration
from .group import GroupConfiguration
from .include import IncludeConfiguration
from .inet_http_server import InetHttpServerConfiguration
from .program import ProgramConfiguration
from .rpcinterface import RpcInterfaceConfiguration
from .supervisorctl import SupervisorctlConfiguration
from .supervisord import SupervisordConfiguration
from .unix_http_server import UnixHttpServerConfiguration

__all__ = (
    "SupervisorConfiguration",
    "SupervisorAirflowConfiguration",
    "load_config",
    "load_airflow_config",
)

_log = getLogger(__name__)


class SupervisorConfiguration(BaseModel):
    def to_cfg(self) -> str:
        ret = ""
        if self.unix_http_server:
            ret += self.unix_http_server.to_cfg() + "\n"
        if self.inet_http_server:
            ret += self.inet_http_server.to_cfg() + "\n"
        if self.supervisord:
            ret += self.supervisord.to_cfg() + "\n"
        if self.supervisorctl:
            ret += self.supervisorctl.to_cfg() + "\n"
        if self.include:
            ret += self.include.to_cfg() + "\n"
        for k, v in self.program.items():
            ret += v.to_cfg(key=k) + "\n"
        for k, v in (self.group or {}).items():
            ret += v.to_cfg(key=k) + "\n"
        for k, v in (self.fcgiprogram or {}).items():
            ret += v.to_cfg(key=k) + "\n"
        for k, v in (self.eventlistener or {}).items():
            ret += v.to_cfg(key=k) + "\n"
        for k, v in (self.rpcinterface or {}).items():
            ret += v.to_cfg(key=k) + "\n"
        return ret

    # supervisor setup
    unix_http_server: Optional[UnixHttpServerConfiguration] = Field(default=None)
    inet_http_server: Optional[InetHttpServerConfiguration] = Field(default=None)
    supervisord: SupervisordConfiguration = Field(default=SupervisordConfiguration())
    supervisorctl: SupervisorctlConfiguration = Field(default=SupervisorctlConfiguration())
    include: Optional[IncludeConfiguration] = Field(default=None)

    program: Dict[str, ProgramConfiguration]
    group: Optional[Dict[str, GroupConfiguration]] = Field(default=None)

    fcgiprogram: Optional[Dict[str, FcgiProgramConfiguration]] = Field(default=None)
    eventlistener: Optional[Dict[str, EventListenerConfiguration]] = Field(default=None)
    rpcinterface: Optional[Dict[str, RpcInterfaceConfiguration]] = Field(default=None)

    # other configuration
    config_path: Optional[Path] = Field(default="", description="Path to supervisor configuration file")
    working_dir: Optional[Path] = Field(default="", description="Path to supervisor working directory")

    @model_validator(mode="after")
    def _setup_config_and_working_dir(self):
        now = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S")
        tempdir = Path(gettempdir()).resolve()

        if self.working_dir == "":
            if self.supervisord.directory:
                # use this as the dir
                self.working_dir = self.supervisord.directory
            else:
                self.working_dir = tempdir / f"supervisor-{now}"
                self.supervisord.directory = self.working_dir
            using_default_working_dir = True
        else:
            using_default_working_dir = False

        # force pidfile to be in working dir if not otherwise set
        if not self.supervisord.pidfile:
            self.supervisord.pidfile = self.working_dir / "supervisord.pid"

        # force logfile to be in working dir if not otherwise set
        if not self.supervisord.logfile:
            self.supervisord.logfile = self.working_dir / "supervisord.log"

        if self.config_path == "":
            if using_default_working_dir:
                self.config_path = self.working_dir / "supervisor.cfg"
            else:
                self.config_path = self.working_dir / f"supervisor-{now}.cfg"

        for name, program_config in self.program.items():
            if program_config.directory is None:
                program_config.directory = self.working_dir / name
        return self

    @classmethod
    def _find_parent_config_folder(
        cls, config_dir: str = "config", config_name: str = "", *, basepath: str = "", _offset: int = 2
    ):
        if basepath:
            if basepath.endswith((".py", ".cfg", ".yml", ".yaml")):
                calling_dag = Path(basepath)
            else:
                calling_dag = Path(basepath) / "dummy.py"
        else:
            calling_dag = Path(_get_calling_dag(offset=_offset))
        folder = calling_dag.parent.resolve()
        exists = (
            (folder / config_dir).exists()
            if not config_name
            else (
                (folder / config_dir / f"{config_name}.yml").exists()
                or (folder / config_dir / f"{config_name}.yaml").exists()
            )
        )
        while not exists:
            folder = folder.parent
            if str(folder) == os.path.abspath(os.sep):
                raise ConfigNotFoundError(config_dir=config_dir, dagfile=calling_dag)
            exists = (
                (folder / config_dir).exists()
                if not config_name
                else (
                    (folder / config_dir / f"{config_name}.yml").exists()
                    or (folder / config_dir / f"{config_name}.yaml").exists()
                )
            )

        config_dir = (folder / config_dir).resolve()
        if not config_name:
            return folder.resolve(), config_dir, ""
        elif (folder / config_dir / f"{config_name}.yml").exists():
            return folder.resolve(), config_dir, (folder / config_dir / f"{config_name}.yml").resolve()
        return folder.resolve(), config_dir, (folder / config_dir / f"{config_name}.yaml").resolve()

    @classmethod
    def load(
        cls: "SupervisorConfiguration",
        config_dir: str = "config",
        config_name: str = "",
        overrides: Optional[list[str]] = None,
        *,
        basepath: str = "",
        _offset: int = 3,
    ) -> "SupervisorConfiguration":
        overrides = overrides or []

        with initialize_config_dir(config_dir=str(Path(__file__).resolve().parent / "hydra"), version_base=None):
            if config_dir:
                hydra_folder, config_dir, _ = cls._find_parent_config_folder(
                    config_dir=config_dir, config_name=config_name, basepath=basepath, _offset=_offset
                )

                cfg = compose(config_name="base", overrides=[], return_hydra_config=True)
                searchpaths = cfg["hydra"]["searchpath"]
                searchpaths.extend([hydra_folder, config_dir])
                if config_name:
                    overrides = [
                        f"+config={config_name}",
                        *overrides.copy(),
                        f"hydra.searchpath=[{','.join(searchpaths)}]",
                    ]
                else:
                    overrides = [*overrides.copy(), f"hydra.searchpath=[{','.join(searchpaths)}]"]

            cfg = compose(config_name="base", overrides=overrides)
            config = instantiate(cfg)

            if not isinstance(config, cls):
                if issubclass(cls, type(config)):
                    config = config.model_dump()
                config = cls(**config)
            return config

    def write(self):
        _log.info(f"Making working dir: {self.working_dir}")
        self.working_dir.mkdir(parents=True, exist_ok=True)
        for program_name, program_config in self.program.items():
            if program_config.directory:
                _log.info(f"Making program dir ({program_name}): {program_config.directory}")
                program_config.directory.mkdir(exist_ok=True)
        _log.info(f"Writing config file: {self.config_path}")
        self.config_path.write_text(self.to_cfg())

    def rmdir(self):
        if not self.running():
            _log.info(f"Removing working dir: {self.working_dir}")
            rmtree(self.working_dir)

    def start(self, daemon: bool = False):
        _log.info(f"Starting supervisord: {self.config_path}")
        if not self.running():
            if daemon is False:
                Popen(f"supervisord -n -c {str(self.config_path)}", shell=True)
                return
            Popen(f"supervisord -c {str(self.config_path)}", close_fds=True, shell=True)

    def running(self):
        # grab the pidfile, find the process with the pid, and kill
        _log.info(f"Checking supervisord: {self.supervisord.pidfile}")
        if not self.supervisord.pidfile.exists():
            return False
        try:
            os.kill(int(self.supervisord.pidfile.read_text()), 0)
        except OSError:
            return False
        return True

    def stop(self):
        if self.running():
            # grab the pidfile, find the process with the pid, and kill with SIGTERM
            _log.info(f"Stopping supervisord: {self.supervisord.pidfile}")
            os.kill(int(self.supervisord.pidfile.read_text()), SIGTERM)

    def kill(self):
        if self.running():
            # grab the pidfile, find the process with the pid, and kill with SIGKILL
            _log.info(f"Killing supervisord: {self.supervisord.pidfile}")
            os.kill(int(self.supervisord.pidfile.read_text()), SIGKILL)


class SupervisorAirflowConfiguration(SupervisorConfiguration):
    airflow: AirflowConfiguration = Field(
        default_factory=AirflowConfiguration, description="Required configurations for Airflow integration"
    )
    check_interval: timedelta = Field(
        default=timedelta(seconds=5), description="Interval between supervisor program status checks"
    )
    check_timeout: timedelta = Field(
        default=timedelta(hours=8), description="Timeout to wait for supervisor program status checks"
    )

    _pydantic_path: Path = PrivateAttr(default="pydantic.json")

    def _write_self(self):
        # TODO make config driven
        self.write()
        _log.info(f"Writing model json: {self._pydantic_path}")
        Path(self._pydantic_path).write_text(self.model_dump_json())

    @model_validator(mode="after")
    def _setup_airflow_defaults(self):
        """Method to overload configuration with values needed for the setup
        of airflow tasks that we construct"""
        # inet_http_server
        if not self.inet_http_server:
            self.inet_http_server = InetHttpServerConfiguration()

        self.inet_http_server.port = self.airflow.port
        self.inet_http_server.username = self.airflow.username
        self.inet_http_server.password = self.airflow.password

        self.supervisorctl.serverurl = (
            f"{self.airflow.protocol}://{self.airflow.host}:{self.airflow.port.split(':')[-1]}/"
        )

        # rpcinterface
        if not self.rpcinterface:
            self.rpcinterface = {"supervisor": RpcInterfaceConfiguration()}
        self.rpcinterface["supervisor"].rpcinterface_factory = self.airflow.rpcinterface_factory

        # supervisord
        self.supervisord.nodaemon = True
        self.supervisord.identifier = "supervisor"

        # programs
        for config in self.program.values():
            config.autostart = False
            config.autorestart = False
            config.startsecs = self.airflow.startsecs
            config.startretries = self.airflow.startretries
            config.exitcodes = self.airflow.exitcodes
            config.stopsignal = self.airflow.stopsignal
            config.stopwaitsecs = self.airflow.stopwaitsecs

        # other
        if str(self.working_dir) not in str(self._pydantic_path):
            self._pydantic_path = self.working_dir / "pydantic.json"
        return self


load_config = SupervisorConfiguration.load
load_airflow_config = SupervisorAirflowConfiguration.load
